<?php

namespace Drupal\entity_template_ui\Form;

use Drupal\Core\Entity\EntityForm;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\Core\TypedData\TypedDataManagerInterface;
use Drupal\Core\Url;
use Drupal\entity_template\Entity\TemplateBlueprint;
use Drupal\entity_template\BlueprintEntityStorageAdaptor;
use Drupal\entity_template\BlueprintEntityAdaptor;
use Drupal\entity_template\BlueprintTempstoreRepository;
use Drupal\entity_template\TemplateBuilderManager;
use Drupal\entity_template\TemplateManager;
use Drupal\entity_template\TemplateBlueprintProviderManager;
use Symfony\Component\DependencyInjection\ContainerInterface;

class TemplateBlueprintForm extends EntityForm {

  /**
   * @var \Drupal\entity_template\TemplateBlueprintProviderManager
   */
  protected $blueprintProviderManager;

  /**
   * @var \Drupal\entity_template\TemplateManager
   */
  protected $templateManager;

  /**
   * @var \Drupal\entity_template\BlueprintTempstoreRepository
   */
  protected $tempstoreRepository;

  /**
   * @var \Drupal\entity_template\BlueprintStorageInterface
   */
  protected $blueprintStorage;

  /**
   * @var \Drupal\Core\TypedData\TypedDataManagerInterface
   */
  protected $typedDataManager;

  /**
   * @var \Drupal\entity_template\TemplateBuilderManager
   */
  protected $templateBuilderManager;

  /**
   * @param \Symfony\Component\DependencyInjection\ContainerInterface $container
   *
   * @return \Drupal\Core\Entity\EntityForm
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('plugin.manager.entity_template.blueprint_provider'),
      $container->get('plugin.manager.entity_template.template'),
      $container->get('entity_template.blueprint_tempstore_repository'),
      $container->get('typed_data_manager'),
      $container->get('plugin.manager.entity_template.builder')
    );
  }

  /**
   * TemplateBlueprintForm constructor.
   *
   * @param \Drupal\entity_template\TemplateManager $template_manager
   */
  public function __construct(
    TemplateBlueprintProviderManager $blueprint_provider_manager,
    TemplateManager $template_manager,
    BlueprintTempstoreRepository $tempstore_repository,
    TypedDataManagerInterface $typed_data_manager,
    TemplateBuilderManager $template_builder_manager
  ) {
    $this->blueprintProviderManager = $blueprint_provider_manager;
    $this->templateManager = $template_manager;
    $this->tempstoreRepository = $tempstore_repository;
    $this->typedDataManager = $typed_data_manager;
    $this->templateBuilderManager = $template_builder_manager;
  }

  /**
   * {@inheritdoc}
   */
  protected function actions(array $form, FormStateInterface $form_state) {
    if (!$this->entity->get('builder')) {
      return [
        'select_submit' => [
          '#type' => 'submit',
          '#value' => $this->t('Select Builder'),
          '#submit' => ['::submitFormSelectBuilder'],
        ],
      ];
    }

    return parent::actions($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function setEntity(EntityInterface $entity) {
    if (!($entity instanceof TemplateBlueprint)) {
      throw new \InvalidArgumentException('This form can only be used with TemplateBlueprint config entities.');
    }

    // We want to check the tempstore here, to do so requires working out
    // whether this is a default blueprint or so that we can correctly set the
    // provider.
    $provider_id = $entity->get('is_default') ? 'builder' : 'blueprint_config';
    /** @var \Drupal\entity_template\Plugin\EntityTemplate\BlueprintProvider\BlueprintProviderInterface $provider */
    $provider = $this->blueprintProviderManager->createInstance($provider_id);
    $blueprint_storage = new BlueprintEntityStorageAdaptor($entity, $provider);

    if ($this->tempstoreRepository->has($blueprint_storage)) {
      $blueprint_storage = $this->tempstoreRepository->get($blueprint_storage);
    }
    else {
      $this->tempstoreRepository->set($blueprint_storage);
    }

    $this->blueprintStorage = $blueprint_storage;
    if ($blueprint_storage instanceof BlueprintEntityAdaptor) {
      $entity = $blueprint_storage->getEntity();
    }

    return parent::setEntity($entity);
  }

  /**
   * @param array $form
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *
   * @return array
   */
  public function form(array $form, FormStateInterface $form_state) {
    $form = parent::form($form, $form_state);
    $form['#tree'] = TRUE;

    $form['label'] = [
      '#title' => $this->t('Label'),
      '#type' => 'textfield',
      '#default_value' => $this->entity->label(),
      '#description' => t('The blueprint label.'),
      '#required' => TRUE,
      '#size' => 30,
    ];
    $form['id'] = [
      '#type' => 'machine_name',
      '#default_value' => $this->entity->id(),
      '#disabled' => $this->entity->id(),
      '#maxlength' => 128,
      '#machine_name' => [
        'exists' => TemplateBlueprint::class.'::load',
        'replace_pattern' => '[^a-z0-9_\.]+',
        'source' => ['label'],
      ],
    ];

    if (!$this->entity->get('builder')) {
      $builder_options = [];
      foreach ($this->templateBuilderManager->getDefinitions() as $key => $definition) {
        $builder_options[$key] = $definition['label'];
      }
      $form['builder'] = [
        '#type' => 'select',
        '#title' => $this->t('Builder'),
        '#options' => $builder_options,
        '#required' => TRUE,
      ];

      return $form;
    }

    $form['contexts_wrapper'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Contexts'),
      '#description' => $this->t('Configure additional contexts required for this blueprint.'),
    ];
    $form['contexts_wrapper']['contexts'] = [
      '#type' => 'table',
      '#header' => [
        $this->t('Context'),
        $this->t('Type'),
        $this->t('Provider'),
        $this->t('Operations'),
      ],
      '#empty' => $this->t('This blueprint has no contexts'),
    ];
    foreach ($this->blueprintStorage->getContextDefinitions() as $name => $definition) {
      if ($this->blueprintStorage->getBuilder()->hasContextDefinition($name)) {
        $provider = $this->t("Builder");
      }
      else if ($name === "current_user") {
        $provider = '';
      }
      else {
        // @todo: Provider becomes the relationship or static context plugin used.
        $provider = '';
      }

      $data_type_definition = $definition->isMultiple() ?
        $this->typedDataManager->createListDataDefinition($definition->getDataType()) :
        $this->typedDataManager->createDataDefinition($definition->getDataType());

      $row = [];
      $row['context'] = [
        '#markup' => $this->t('@label <small>[@name]</small>', [
          '@label' => $definition->getLabel(),
          '@name' => $name,
        ])
      ];
      $row['type'] = [
        '#markup' => $data_type_definition->getLabel() ?: $this->typedDataManager->getDefinition($data_type_definition->getDataType())['label'],
      ];
      $row['provider'] = [
        '#markup' => $provider,
      ];
      $row['operations'] = [
        '#type' => 'operations',
        '#links' => [],
      ];

      $form['contexts_wrapper']['contexts'][$name] = $row;
    }

    $form['templates_wrapper'] = [
      '#type' => 'fieldset',
      '#title' => $this->t('Templates'),
      '#description' => $this->t('Configure the templates required for this blueprint.'),
    ];
    $form['templates_wrapper']['add'] = [
      '#type' => 'link',
      '#title' => $this->t('Add Template'),
      '#url' => Url::fromRoute(
        'entity_template_ui.choose_template',
        [
          'template_blueprint_provider' => $this->blueprintStorage->getProvider()->getPluginId(),
          'blueprint_storage' => $this->blueprintStorage->getProviderKey(),
        ],
        [
          'attributes' => [
            'class' => [
              'use-ajax',
            ],
            'data-dialog-type' => 'dialog',
            'data-dialog-renderer' => 'off_canvas',
          ],
        ]
      ),
      '#weight' => -5,
      '#attributes' => [
        'class' => ['add-template-button'],
      ],
    ];
    $form['templates_wrapper']['templates'] = [
      '#type' => 'table',
      '#parents' => ['templates'],
      '#header' => [
        [
          'data' => $this->t('Template'),
          'colspan' => 2,
        ],
        $this->t('Applies When'),
        $this->t('Builds'),
        $this->t('Operations'),
      ],
      '#tabledrag' => [
        [
          'action' => 'order',
          'relationship' => 'sibling',
          'group' => 'template-priority',
        ]
      ]
    ];

    /** @var \Drupal\entity_template\Plugin\EntityTemplate\Template\TemplateInterface $template */
    foreach ($this->blueprintStorage->getTemplates() as $uuid => $template) {
      $config = $template->getConfiguration();

      /** @var \Drupal\Core\Condition\ConditionInterface[] $conditions */
      if ($conditions = $template->getConditions()) {
        $applies_when = [
          '#theme' => 'item_list',
          '#items' => [],
        ];

        foreach ($conditions as $condition) {
          $applies_when['#items'][] = $condition->summary();
        }
      }
      else {
        $applies_when = [ '#markup' => new TranslatableMarkup('Always'), ];
      }

      $row = [];
      $row['#attributes']['class'][] = 'draggable';
      $row['label'] = [
        '#type' => 'container',
        'label' => [
          '#type' => 'html_tag',
          '#tag' => 'h4',
          '#value' => $template->label(),
          '#attributes' => [
            'class' => ['template-label'],
          ],
        ],
        'description' => [
          '#type' => 'html_tag',
          '#tag' => 'p',
          '#value' => isset($config['description']) ? $config['description'] : '',
          '#attributes' => [
            'class' => ['template-description'],
          ],
          '#access' => !empty($config['description']),
        ],
      ];
      $row['priority'] = [
        '#type' => 'textfield',
        '#title' => $this->t('Priority'),
        '#title_display' => 'invisible',
        '#default_value' => 0 - $config['priority'],
        '#size' => 3,
        '#attributes' => [
          'class' => [ 'template-priority' ],
        ],
      ];
      $row['applies_when'] = $applies_when;
      $row['builds'] = [
        '#markup' => $template->getReturnTypeLabel(),
      ];
      $row['operations'] = [
        '#type' => 'dropbutton',
        '#links' => [
          'configure' => [
            'title' => $this->t('configure'),
            'url' => Url::fromRoute(
              'entity_template_ui.edit_template',
              [
                'template_blueprint_provider' => $this->blueprintStorage->getProvider()->getPluginId(),
                'blueprint_storage' => $this->blueprintStorage->getProviderKey(),
                'template_id' => $uuid,
              ],
              [
                'query' => $this->getDestinationArray(),
              ]
            ),
          ],
          'remove' => [
            'title' => $this->t('remove'),
            'url' => Url::fromRoute(
              'entity_template_ui.remove_template',
              [
                'template_blueprint_provider' => $this->blueprintStorage->getProvider()->getPluginId(),
                'blueprint_storage' => $this->blueprintStorage->getProviderKey(),
                'template_id' => $uuid,
              ]
            )
          ]
        ]

      ];

      $form['templates_wrapper']['templates'][$uuid] = $row;
    }

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  protected function copyFormValuesToEntity(EntityInterface $entity, array $form, FormStateInterface $form_state) {
    parent::copyFormValuesToEntity($entity, $form, $form_state);

    $entity->set('templates', $this->blueprintStorage->getTemplates()->getConfiguration());
  }

  /**
   * {@inheritdoc}
   */
  public function save(array $form, FormStateInterface $form_state) {
    $return = parent::save($form, $form_state);

    $this->tempstoreRepository->delete($this->blueprintStorage);

    return $return;
  }

  /**
   * Submit callback for selecting a builder.
   *
   * @param array $form
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *
   * @throws \Drupal\Core\TempStore\TempStoreException
   */
  public function submitFormSelectBuilder(array &$form, FormStateInterface $form_state) {
    $this->blueprintStorage->getEntity()->set('builder', $form_state->getValue('builder'));
    $this->tempstoreRepository->set($this->blueprintStorage);

    $form_state->setRebuild();
  }

}
