<?php

namespace Drupal\entity_template;

use Drupal\Core\DependencyInjection\DependencySerializationTrait;
use Drupal\Core\Plugin\Context\Context;
use Drupal\Core\Plugin\Context\ContextInterface;
use Drupal\Core\Plugin\Context\EntityContextDefinition;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\entity_template\Plugin\EntityTemplate\Builder\BuilderInterface;

/**
 * Class Blueprint
 *
 * @package Drupal\entity_template
 */
class Blueprint implements BlueprintInterface {
  use SampleContextValueTrait;
  use DependencySerializationTrait;

  /**
   * The entity type manager service.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The Builder.
   *
   * @var \Drupal\entity_template\Plugin\EntityTemplate\Builder\BuilderInterface
   */
  protected $builder;

  /**
   * Template configuration.
   *
   * @var array
   */
  protected $templates;

  /**
   * Priority
   *
   * @var integer
   */
  protected $priority = 0;

  /**
   * @var \Drupal\Core\Plugin\DefaultLazyPluginCollection
   */
  protected $templatesCollection;

  /**
   * Contexts by
   *
   * @var \Drupal\Core\Plugin\Context\ContextInterface[]
   */
  protected $contexts = [];

  /**
   * Context definitions.
   *
   * @var \Drupal\Core\Plugin\Context\ContextDefinitionInterface[]
   */
  protected $contextDefinitions = [];

  /**
   * Blueprint constructor.
   *
   * @param \Drupal\entity_template\Plugin\EntityTemplate\Builder\BuilderInterface $builder
   * @param array $templates
   */
  public function __construct(BuilderInterface $builder, array $templates = []) {
    $this->builder = $builder;
    $this->templates = $templates;
  }

  /**
   * {@inheritdoc}
   */
  public function getContexts($sample = FALSE) {
    $contexts = [];
    foreach ($this->getContextDefinitions() as $name => $definition) {
      $contexts[$name] = $this->getContext($name, $sample);
    }
    return $contexts;
  }

  /**
   * {@inheritdoc}
   */
  public function getContext($name, $sample = FALSE) {
    if (!isset($this->contexts[$name]) && ($def = $this->getContextDefinition($name))) {
      if ($def->getDefaultValue()) {
        $this->contexts[$name] = new Context($def, $def->getDefaultValue());
      }
      else if ($sample) {
        $this->contexts[$name] = new Context($def, $this->generateSampleValue($def));
      }
    }

    if (!isset($this->contexts[$name])) {
      throw new \InvalidArgumentException("No context {$name} available.");
    }

    return $this->contexts[$name];
  }

  /**
   * {@inheritdoc}
   */
  public function setContext($name, ContextInterface $context) {
    if (!$this->getContextDefinition($name) || $this->getContextDefinition($name)->getDataType() != $context->getContextDefinition()->getDataType()) {
      throw new \InvalidArgumentException('Invalid context supplied for '.$name);
    }

    $this->contexts[$name] = $context;

    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function clearContexts() {
    $this->contexts = [];
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getContextValues($sample = FALSE) {
    $values = [];

    foreach ($this->getContextDefinitions() as $name => $definition) {
      $values[$name] = $this->getContextValue($name, $sample);
    }

    return $values;
  }

  /**
   * {@inheritdoc}
   */
  public function getContextValue($name, $sample = FALSE) {
    return $this->getContext($name, $sample)->getContextValue();
  }

  /**
   * {@inheritdoc}
   */
  public function getContextDefinitions() {
    if (empty($this->contextDefinitions)) {
      $this->contextDefinitions = $this->getBuilder()->getContextDefinitions();
      $this->contextDefinitions += $this->getExtraContextDefinitions();
    }

    return $this->contextDefinitions;
  }

  /**
   * {@inheritdoc}
   */
  public function getContextDefinition($name) {
    $definitions = $this->getContextDefinitions();
    if (isset($definitions[$name])) {
      return $definitions[$name];
    }

    throw new \InvalidArgumentException('No context definition found for '.$name);
  }

  /**
   * Get extra context definitions.
   *
   * @return \Drupal\Core\Plugin\Context\ContextDefinitionInterface[]
   */
  protected function getExtraContextDefinitions() {
    $current_user_context = new EntityContextDefinition('user', new TranslatableMarkup('Current User'));
    $current_user_context->setDefaultValue(
      \Drupal::entityTypeManager()
        ->getStorage('user')
        ->load(\Drupal::currentUser()->id())
    );

    return [
      'current_user' => $current_user_context,
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function getPriority() {
    return $this->priority;
  }

  /**
   * {@inheritdoc}
   */
  public function setPriority($priority) {
    $this->priority = $priority;
    return $this;
  }

  /**
   * Get the builder this blueprint satisfies.
   *
   * @return \Drupal\entity_template\Plugin\EntityTemplate\Builder\BuilderInterface
   */
  public function getBuilder() {
    return $this->builder;
  }

  /**
   * Execute the blueprint.
   *
   * @param array $parameters
   *
   * @return \Drupal\entity_template\BlueprintResult
   */
  public function execute(array $parameters = []) {
    // Set up contexts ready for execution.
    foreach ($parameters as $name => $value) {
      // Just ignore extra parameters.
      if (!$this->getContextDefinition($name)) {
        continue;
      }

      if ($value instanceof ContextInterface) {
        $this->setContext($name, $value);
      }
      else {
        $this->setContext(
          $name,
          new Context($this->getContextDefinition($name), $value)
        );
      }
    }
    $this->prepareContexts();

    $result = new BlueprintResult();
    foreach ($this->getTemplates() as $key => $template) {
      /** @var \Drupal\entity_template\Plugin\EntityTemplate\Template\TemplateInterface $template */
      $template->setBlueprintResult($result);

      if (!$template->applies()) {
        continue;
      }

      $template_result = $template->execute();

      foreach ($template_result->getMessages() as $message) {
        $result->addMessage($key, $message);
      }

      if ($template->isSingle() || !($template_result instanceof MultipleTemplateResult)) {
        if ($entity = $template_result->getEntity()) {
          $result->addItem($key, $entity);

          // If the builder doesn't return a list then we stop after the first
          // successful generation.
          if (!$this->getBuilder()->returnsList()) {
            break;
          }
        }
      }
      else {
        /** @var MultipleTemplateResult $template_result */
        foreach ($template_result->getEntities() as $sub_key => $entity) {
          $result->addItem("{$key}:{$sub_key}", $entity);
        }
      }
    }

    return $result;
  }

  /**
   * Prepare the contexts before the blueprint is executed.
   */
  protected function prepareContexts() {}

  /**
   * {@inheritdoc}
   */
  public function getTemplates() {
    if (!$this->templatesCollection) {
      $this->templatesCollection = new BlueprintLazyTemplatesCollection(
        \Drupal::service('plugin.manager.entity_template.template'),
        $this,
        $this->getTemplatesConfiguration()
      );
    }

    return $this->templatesCollection;
  }

  /**
   * {@inheritdoc}
   */
  public function getTemplate($id) {
    return $this->getTemplates()->get($id);
  }

  /**
   * {@inheritdoc}
   */
  public function getTemplatesConfiguration() {
    if (!$this->templatesCollection) {
      return $this->templates;
    }

    return $this->templatesCollection->getConfiguration();
  }

}
