<?php

namespace Drupal\entity_template\Plugin\Derivative;

use Drupal\Component\Plugin\Derivative\DeriverBase;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Entity\FieldableEntityInterface;
use Drupal\Core\Field\BaseFieldDefinition;
use Drupal\Core\Field\FieldTypePluginManagerInterface;
use Drupal\Core\Plugin\Discovery\ContainerDeriverInterface;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Class EntityFieldStorageDeriver
 *
 * @package Drupal\entity_template\Plugin\Derivative
 */
class EntityFieldStorageDeriver extends DeriverBase implements ContainerDeriverInterface {

  /**
   * @var string
   */
  protected $basePluginId;

  /**
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface
   */
  protected $entityFieldManager;

  /**
   * @var \Drupal\Core\Field\FieldTypePluginManagerInterface
   */
  protected $fieldTypeManager;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, $base_plugin_id) {
    return new static(
      $base_plugin_id,
      $container->get('entity_type.manager'),
      $container->get('entity_field.manager'),
      $container->get('plugin.manager.field.field_type')
    );
  }

  /**
   * EntityFieldStorageDeriver constructor.
   *
   * @param string $base_plugin_id
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   * @param \Drupal\Core\Entity\EntityFieldManagerInterface $entity_field_manager
   */
  public function __construct(
    $base_plugin_id,
    EntityTypeManagerInterface $entity_type_manager,
    EntityFieldManagerInterface $entity_field_manager,
    FieldTypePluginManagerInterface $field_type_manager
  ) {
    $this->basePluginId = $base_plugin_id;
    $this->entityTypeManager = $entity_type_manager;
    $this->entityFieldManager = $entity_field_manager;
    $this->fieldTypeManager = $field_type_manager;
  }

  /**
   * {@inheritdoc}
   */
  public function getDerivativeDefinitions($base_plugin_definition) {
    $field_map = $this->entityFieldManager->getFieldMap();

    foreach ($this->entityTypeManager->getDefinitions() as $entity_type_id => $entity_type) {
      // Only works on fieldable entities.
      if (!is_subclass_of($entity_type->getClass(), FieldableEntityInterface::class)) {
        continue;
      }

      foreach ($this->entityFieldManager->getFieldStorageDefinitions($entity_type_id) as $field_name => $definition) {
        if (!isset($field_map[$entity_type_id][$field_name])) {
          continue;
        }

        if (!empty($base_plugin_definition['field_types'])) {
          $field_types = is_array($base_plugin_definition['field_types']) ? $base_plugin_definition['field_types'] : [$base_plugin_definition['field_types']];
          if (!in_array($definition->getType(), $field_types)) {
            continue;
          }
        }

        foreach ($field_map[$entity_type_id][$field_name]['bundles'] as $bundle) {
          $class = $base_plugin_definition['class'];
          $field_type = $this->fieldTypeManager->getDefinition($definition->getType());
          if (
            isset($field_type['entity_template_component'][$this->basePluginId]) &&
            is_subclass_of($field_type['entity_template_component'][$this->basePluginId], $class)
          ) {
            $class = $field_type['entity_template_component'][$this->basePluginId];
          }

          $id = "{$entity_type_id}.{$field_name}";
          if ($entity_type->hasKey('bundle') && !$definition->isBaseField()) {
            $applies_to = ["entity:{$entity_type_id}:{$bundle}"];
            $id .= ".{$bundle}";
          }
          else {
            $applies_to = ["entity:{$entity_type_id}"];
          }
          $this->derivatives[$id] = [
              "class" => $class,
              "label" => new TranslatableMarkup("@label (@type)", [
                '@label' => $definition->getLabel(),
                '@type' => $base_plugin_definition['label'],
              ]),
              "entity_type_id" => $entity_type_id,
              "bundle" => $bundle,
              "field_name" => $field_name,
              "applies_to" => $applies_to,
            ] + $base_plugin_definition;
        }
      }
    }

    return parent::getDerivativeDefinitions($base_plugin_definition);
  }
}
