<?php

namespace Drupal\entity_template\Plugin\EntityTemplate\Component;

use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Field\FieldTypePluginManagerInterface;
use Drupal\Core\Field\WidgetPluginManager;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Form\SubformState;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\entity_template\TemplateResult;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Class FieldWidgetInputComponent
 *
 * @EntityTemplateComponent(
 *   id = "field.widget_input",
 *   label = "Input Value",
 *   category = "Direct Input",
 *   deriver = "\Drupal\entity_template\Plugin\Derivative\EntityFieldStorageDeriver",
 * )
 *
 * @package Drupal\entity_template\Plugin\EntityTemplate\Component
 */
class FieldWidgetInputComponent extends FieldComponentBase implements SwappableComponentInterface {

  /**
   * @var \Drupal\Core\Field\WidgetPluginManager
   */
  protected $widgetManager;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('entity_type.manager'),
      $container->get('entity_field.manager'),
      $container->get('plugin.manager.field.field_type'),
      $container->get('plugin.manager.field.widget')
    );
  }

  /**
   * FieldWidgetInputComponent constructor.
   *
   * @param array $configuration
   * @param $plugin_id
   * @param $plugin_definition
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   * @param \Drupal\Core\Entity\EntityFieldManagerInterface $entity_field_manager
   * @param \Drupal\Core\Field\FieldTypePluginManagerInterface $field_type_manager
   * @param \Drupal\Core\Field\WidgetPluginManager $widget_manager
   */
  public function __construct(
    array $configuration,
    $plugin_id,
    $plugin_definition,
    EntityTypeManagerInterface $entity_type_manager,
    EntityFieldManagerInterface $entity_field_manager,
    FieldTypePluginManagerInterface $field_type_manager,
    WidgetPluginManager $widget_manager
  ) {
    parent::__construct($configuration, $plugin_id, $plugin_definition, $entity_type_manager, $entity_field_manager, $field_type_manager);

    $this->widgetManager = $widget_manager;
  }

  /**
   * The value input.
   *
   * @return mixed
   */
  public function getValue() {
    return isset($this->configuration['value']) ? $this->configuration['value'] : [];
  }

  /**
   * {@inheritdoc}
   */
  public function apply(EntityInterface $entity, TemplateResult $result) {
    $entity->{$this->getFieldName()} = $this->getValue();
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $field_definition = $this->getFieldDefinition();

    // Get class
    $widget = $this->getWidget($field_definition);
    $items = $this->fieldTypeManager->createFieldItemList($this->getSampleEntity(), $this->getFieldName(), $this->getValue());
    $form[$field_definition->getName()] = $widget->form($items,$form, $form_state);

    return parent::buildConfigurationForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    $field_definition = $this->getFieldDefinition();

    $widget = $this->getWidget($field_definition);
    $items = $this->fieldTypeManager->createFieldItemList($this->getSampleEntity(), $this->getFieldName(), []);
    $widget->extractFormValues($items, $form, ($form_state instanceof SubformState) ? $form_state->getCompleteFormState() : $form_state);

    $this->configuration['value'] = $items->getValue();
    parent::submitConfigurationForm($form, $form_state);
  }

  /**
   * Get the widget for this field definition.
   *
   * @param \Drupal\Core\Field\FieldDefinitionInterface $field_definition
   *
   * @return \Drupal\Core\Field\WidgetInterface
   *
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  protected function getWidget(FieldDefinitionInterface $field_definition) {
    if (($display = $field_definition->getDisplayOptions('form')) && isset($display['type'])) {
      $widget = $this->widgetManager->createInstance(
        $display['type'],
        [
          'field_definition' => $field_definition,
        ] + $display + [
          'settings' => [],
          'third_party_settings' => [],
        ]
      );
    }
    else {
      $field_type_definition = $this->fieldTypeManager->getDefinition($field_definition->getType());

      if (isset($field_type_definition['default_widget'])) {
        $default_widget = $field_type_definition['default_widget'];
      }
      else {
        $options = $this->widgetManager->getOptions($field_definition->getType());
        $default_widget = key($options);
      }

      $widget = $this->widgetManager->createInstance($default_widget, [
        'field_definition' => $field_definition,
        'settings' => [],
        'third_party_settings' => [],
      ]);
    }

    return $widget;
  }

  /**
   * Compute whether the component is swappable.
   *
   * @return bool
   */
  public function isSwappable(): bool {
    return !empty($this->canSwapTo());
  }

  /**
   * Get a list of plugin ids that can be swapped between.
   *
   * @return array
   */
  public function canSwapTo(): array {
    $entity_type_id = $this->getEntityTypeId();
    $field_name = $this->getFieldName();

    $other = "field.data_select:{$entity_type_id}.{$field_name}";
    if (\Drupal::service('plugin.manager.entity_template.component')->hasDefinition($other)) {
      return [
        $other => new TranslatableMarkup('Value Select'),
      ];
    }

    return [];
  }

  /**
   * {@inheritdoc}
   */
  public function swapConfigurationTo(ComponentInterface $component): ComponentInterface {
    return $component;
  }

}
