<?php

namespace Drupal\entity_template\Plugin\EntityTemplate\Template;

use Drupal\Component\Plugin\Context\ContextInterface as ComponentContextInterface;
use Drupal\Component\Plugin\Exception\ContextException;
use Drupal\Component\Plugin\Exception\PluginException;
use Drupal\Component\Uuid\UuidInterface;
use Drupal\Core\Cache\CacheableDependencyInterface;
use Drupal\Core\Cache\CacheableMetadata;
use Drupal\Core\Condition\ConditionManager;
use Drupal\Core\Entity\EntityTypeBundleInfo;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Plugin\Context\Context;
use Drupal\Core\Plugin\Context\ContextDefinition;
use Drupal\Core\Plugin\Context\ContextHandlerInterface;
use Drupal\Core\Plugin\ContextAwarePluginInterface;
use Drupal\Core\Plugin\DefaultLazyPluginCollection;
use Drupal\entity_template\BlueprintResult;
use Drupal\entity_template\Exception\TemplateException;
use Drupal\entity_template\TemplateComponentManager;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * An entity template. This is takes a blueprint and uses the contexts on it
 * to produce an entity.
 *
 * @EntityTemplate(
 *   id = "default",
 *   label = @Translation("Template"),
 *   single = "TRUE",
 * );
 *
 * @package Drupal\entity_template
 */
class BlueprintTemplate extends Template implements BlueprintTemplateInterface {

  /**
   * @var \Drupal\entity_template\BlueprintInterface
   */
  protected $blueprint;

  /**
   * @var \Drupal\Component\Plugin\LazyPluginCollection
   */
  protected $conditionPluginCollection;

  /**
   * @var \Drupal\Core\Condition\ConditionManager
   */
  protected $conditionManager;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('entity_type.manager'),
      $container->get('entity_type.bundle.info'),
      $container->get('context.handler'),
      $container->get('plugin.manager.condition'),
      $container->get('plugin.manager.entity_template.component'),
      $container->get('uuid')
    );
  }

  /**
   * BlueprintTemplate constructor.
   *
   * @param array $configuration
   * @param $plugin_id
   * @param $plugin_definition
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   * @param \Drupal\Core\Entity\EntityTypeBundleInfo $entity_type_bundle_info
   * @param \Drupal\Core\Plugin\Context\ContextHandlerInterface $context_handler
   * @param \Drupal\Core\Condition\ConditionManager $condition_manager
   * @param \Drupal\entity_template\TemplateComponentManager $component_manager
   * @param \Drupal\Component\Uuid\UuidInterface $uuid_generator
   */
  public function __construct(
    array $configuration,
    $plugin_id,
    $plugin_definition,
    EntityTypeManagerInterface $entity_type_manager,
    EntityTypeBundleInfo $entity_type_bundle_info,
    ContextHandlerInterface $context_handler,
    ConditionManager $condition_manager,
    TemplateComponentManager $component_manager,
    UuidInterface $uuid_generator
  ) {
    $this->conditionManager = $condition_manager;
    $this->blueprint = isset($configuration['blueprint']) ? $configuration['blueprint'] : NULL;
    unset($configuration['blueprint']);

    parent::__construct(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $entity_type_manager,
      $entity_type_bundle_info,
      $context_handler,
      $component_manager,
      $uuid_generator
    );
  }

  /**
   * Get the blueprint.
   *
   * @return \Drupal\entity_template\BlueprintInterface
   */
  public function getBlueprint() {
    return $this->blueprint;
  }

  /**
   * {@inheritdoc}
   */
  public function getContexts($sample = FALSE) {
    return ($this->getBlueprint()->getContexts($sample) ?: []) + ($this->getExtraContexts($sample) ?: []);
  }

  /**
   * {@inheritdoc}
   */
  public function getContext($name, $sample = FALSE) {
    try {
      return $this->getBlueprint()->getContext($name, $sample);
    }
    catch (\InvalidArgumentException $exception) {
      return $this->getExtraContext($name, $sample);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function setContext($name, ComponentContextInterface $context) {
    try {
      if ($this->getBlueprint()->getContextDefinition($name)) {
        $this->getBlueprint()->setContext($name, $context);
      }
    }
    catch (\InvalidArgumentException $exception) {
      return parent::setContext($name, $context);
    }

    return $this;
  }

  /**
   * Get all the context definition interfaces
   *
   * @return \Drupal\Core\Plugin\Context\ContextDefinitionInterface[]
   */
  public function getContextDefinitions() {
    return $this->getBlueprint()->getContextDefinitions() + parent::getContextDefinitions();
  }

  /**
   * {@inheritdoc}
   */
  protected function getExtraContextDefinitions() {
    return [
      '_blueprint_result' => new ContextDefinition(
        'entity_template:blueprint_result',
        'Blueprint Result',
        TRUE
      ),
    ];
  }

  /**
   * @param \Drupal\entity_template\BlueprintResult $result
   *
   * @return static
   */
  public function setBlueprintResult(BlueprintResult $result) {
    $this->setContext(
      '_blueprint_result',
      new Context($this->getContextDefinition('_blueprint_result'))
    );

    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function applies(CacheableMetadata $cache_metadata = NULL) {
    $this->prepareContexts();

    try {
      foreach ($this->getConditions() as $condition) {
        if ($condition instanceof ContextAwarePluginInterface) {
          $this->contextHandler->applyContextMapping($condition, $this->getContexts());
        }

        if ($cache_metadata && $condition instanceof CacheableDependencyInterface) {
          $cache_metadata->addCacheableDependency($condition);
        }

        if (!$condition->evaluate()) {
          return FALSE;
        }
      }
    }
    catch (PluginException $e) {
      return FALSE;
    }
    catch (ContextException $e) {
      return FALSE;
    }

    return TRUE;
  }

  /**
   * {@inheritdoc}
   */
  protected function getTargetEntityTypeId() {
    $return_type = $this->getBlueprint()->getBuilder()->getReturnType();

    // Strip off [] if necessary.
    if (substr($return_type, -2) === '[]') {
      $return_type = substr($return_type, 0, -2);
    }

    // If there is no : then we assume the return type is 'entity'. In this case
    // we need to look in configuration.
    if (!strpos($return_type, ':')) {
      if (isset($this->configuration['target_entity_type_id'])) {
        return $this->configuration['target_entity_type_id'];
      }

      throw new TemplateException("No entity type id available.");
    }
    else {
      list(,$entity_type_id) = explode(':', $return_type);
      return $entity_type_id;
    }
  }

  /**
   * Get the target entity bundle where applicable
   *
   * @return string
   *
   * @throws \Drupal\entity_template\Exception\TemplateException
   */
  protected function getTargetEntityBundle() {
    $return_type = $this->getBlueprint()->getBuilder()->getReturnType();

    // Strip off [] if necessary.
    if (substr($return_type, -2) === '[]') {
      $return_type = substr($return_type, 0, -2);
    }

    if (!strpos($return_type, ':') || count(explode(':', $return_type)) < 3) {
      if (isset($this->configuration['target_entity_bundle'])) {
        return $this->configuration['target_entity_bundle'];
      }

      return NULL;
    }
    else {
      list(,,$bundle) = explode(':', $return_type);
      return $bundle;
    }
  }

  /**
   * Get the conditions of the template.
   *
   * @return \Drupal\Component\Plugin\LazyPluginCollection
   */
  public function getConditions() {
    if (!$this->conditionPluginCollection) {
      $this->conditionPluginCollection = new DefaultLazyPluginCollection(
        $this->conditionManager,
        $this->configuration['conditions'] ?: []
      );
    }

    return $this->conditionPluginCollection;
  }

  /**
   * {@inheritdoc}
   */
  public function addCondition(array $configuration) {
    $configuration['uuid'] = $this->uuidGenerator->generate();
    $this->getConditions()->addInstanceId($configuration['uuid'], $configuration);

    return $configuration['uuid'];
  }

  /**
   * {@inheritdoc}
   */
  public function removeCondition($id) {
    $this->getConditions()->removeInstanceId($id);
  }

  /**
   * {@inheritdoc}
   */
  public function getCondition($id) {
    return $this->getConditions()->get($id);
  }

}
